function [] = Ex_28p2_001_symbolic_algebra()
% ---------------------------------------------------------------------------
% Exercise 28.2 in MacKay's book; by Jingquang Zheng and Matthew Juniper 2022
% ---------------------------------------------------------------------------
% PRIOR READING
%  Information Theory, Inference, and Learning Algorithms, by David J.C.
%  MacKay, Cambridge University Press (v7.2 2005),
%  https://www.inference.org.uk/itprnn/book.pdf 
% 
% Essential reading:
%  Chapters  1 -  3; 
%  Chapters 27 - 28.
%
% Useful reading:
%  Chapters 21, 22, 24.
%  Exercise 28.1
% ---------------------------------------------------------------------------

% ---------------------------------------------------------------------------
% QUESTION
% ---------------------------------------------------------------------------
% We are given data, D, and asked to evaluate the evidence (a.k.a. marginal
% likelihood) for two models, H1(w) and H2(w), given the data, D, where
% w are the models' parameters. In other words:
% 
%   Calculate the evidences, P(D|H1) and P(D|H2), where:
% 
%   P(D|Hi) = int{ P(D|w,Hi) * P(w|Hi) } dw  (Eq. 28.7) 
% 
%   D is [x] = [ -8 , -2 ,  6 ]
%        [t]   [  8 , 10 , 11 ]
%
%   The measurement, t, is y + Normal(0,1)
% 
%   Model H1 is y = w0
%   Model H2 is y = w0 + w1*x
% 
%   Prior w0 is Normal(0,1) for both models
%   Prior w1 is Normal(0,1)

% ---------------------------------------------------------------------------
% SOLUTION WITH SYMBOLIC ALGEBRA
% ---------------------------------------------------------------------------
% This problem can be solved by hand but here we will use symbolic algebra
% in Matlab, which is just as easy to read but can be adapted easily. 
%
% Start the clock
tic
% Define two symbolic variables
syms w0 w1 

%% Model H1
% Set the prior P(w|H1) = Normal(0,1)
P_w_H1   = (2*pi)^(-0.5) * exp( -0.5 * w0^2 );
% Set the likelihood P(D|w,H1)
P_D_w_H1 = (2*pi)^(-1.5) * exp( -0.5 * ( (8-w0)^2 + (10-w0)^2 + (11-w0)^2 ) );
% Calculate the posterior P(w|D,H1) (Eq. 28.4)
P_w_D_H1 = P_w_H1 * P_D_w_H1;
% Calculate P(D|H1), which is the integral of P(w|D,H1) over w0-space 
P_D_H1   = eval(int(P_w_D_H1, w0, -inf,+inf));

%% Model H2
% Set the prior P(w|H1) = Normal(0,1)
P_w_H2   = (2*pi)^(-1.0) * exp( -0.5 * w0^2 ) * exp( -0.5 * w1^2 );
% Set the likelihood P(D|w,H1)
P_D_w_H2 = (2*pi)^(-1.5) * exp( -0.5 * ( ( 8-(w0+w1*(-8)) )^2 + ( 10-(w0+w1*(-2)) )^2 + (11-(w0+w1*(6)))^2 ) );
% Calculate the posterior P(w|D,H1) (Eq. 28.4)
P_w_D_H2 = P_w_H2 * P_D_w_H2;
% Calculate P(D|H1), which is the integral of P(w|D,H1) over (w0,w1)-space 
P_D_H2   = eval(int(int(P_w_D_H2, w0, -inf, +inf), w1, -inf,+inf));

%% Display results to screen
fprintf('------------------------------------------------ \n')
fprintf(' log(evidence) calculated with symbolic algebra \n')
fprintf('------------------------------------------------ \n')
fprintf('log{ P(D|H1) } = %6.4f \n',log(P_D_H1))
fprintf('log{ P(D|H2) } = %6.4f \n',log(P_D_H2))

% ---------------------------------------------------------------------------
% RESULTS AND DISCUSSION
% ---------------------------------------------------------------------------
% The results are:
%
%  log(evidence) calculated with symbolic algebra 
%  ---------------------------------------------- 
%  log{ P(D|H1) } = -40.8250 
%  log{ P(D|H2) } = -42.5335 
%
% The probability ratio between model H1 and model H2 is:
%
%  P(H1|D)   P(H1)   P(D|H1)
%  ------- = ----- * -------         (Eq. 28.1)
%  P(H2|D)   P(H2)   P(D|H2)
%
% The first ratio on the RHS, P(H1)/P(H2), measures how much our initial
% beliefs favoured H1 over H2. Let us presume that we had equal belief
% about each model initially, so that this ratio is 1. 
%
% The second ratio is exp{ log{ P(D|H1) } - log{ P(D|H2) } }. This
% expresses how well the observed data were predicted by H1, compared to
% H2:
fprintf('------------------------------------------------ \n')
fprintf('P(D|H1) / P(D|H2) = %6.4f \n',P_D_H1 / P_D_H2)

% ---------------------------------------------------------------------------
% EXTENSION TO FIND THE MAP POINT
% ---------------------------------------------------------------------------
% Although not asked for in this in the question, we can also find the parameter
% values, w, that maximize the posterior P(w|D,H1) using symbolic algebra. 
%
% P(w|D,Hi) is maximized when its partial derivatives w.r.t. its parameters
% are zero. 

%% Model H1
% Create partial(P(w|D,H1))/partial(w0)
d_P_w_D_H1_dw0 = diff(P_w_D_H1, w0);
% Find w0 that sets this derivative to zero
w0_MAP_H1 = vpa(solve(d_P_w_D_H1_dw0 == 0));

%% Model H2
% Create partial(P(w|D,H2))/partial(w0)
d_P_w_D_H2_dw0 = diff(P_w_D_H2, w0);
% Create partial(P(w|D,H2))/partial(w1)
d_P_w_D_H2_dw1 = diff(P_w_D_H2, w1);
% Find (w0,w1) that sets these derivatives to zero simultaneously 
wi_MAP_H2 = solve([d_P_w_D_H2_dw0 == 0, d_P_w_D_H2_dw1 == 0], [w0 , w1]);
% Extract these from the above structure
w0_MAP_H2 = vpa(wi_MAP_H2.w0);
w1_MAP_H2 = vpa(wi_MAP_H2.w1);

% Stop the clock
time_taken = toc;

%% Display results to screen
fprintf('------------------------------------------------ \n')
fprintf(' MAP point calculated with symbolic algebra      \n')
fprintf('------------------------------------------------ \n')
fprintf('max{ P(D|H1) } occurs at w = ( %6.4f )   \n',w0_MAP_H1)
fprintf('max{ P(D|H2) } occurs at w = ( %6.4f , %6.4f ) \n',w0_MAP_H2,w1_MAP_H2)

fprintf('------------------------------------------------ \n')
fprintf('time taken for symbolic algebra = %6.4f secs \n',time_taken)

end
