function [] = Ex_28p2_003_extension()
% ---------------------------------------------------------------------------
% Extension to exercise 28.2 in MacKay's book; by Matthew Juniper 2021
% ---------------------------------------------------------------------------
% PRIOR READING
%  Information Theory, Inference, and Learning Algorithms, by David J.C.
%  MacKay, Cambridge University Press (v7.2 2005),
%  https://www.inference.org.uk/itprnn/book.pdf 
% 
% Essential reading:
%  Chapters  1 -  3; 
%  Chapters 27 - 28.
%
% Useful reading:
%  Chapters 21, 22, 24.
%  Exercise 28.1
% ---------------------------------------------------------------------------

% ---------------------------------------------------------------------------
% QUESTION WITH EXTENSIONS
% ---------------------------------------------------------------------------
% We are given data, D, and asked to evaluate the evidence (a.k.a. marginal
% likelihood) for two models, H1(w) and H2(w), given the data, D, where
% w are the models' parameters. In other words:
% 
%   Calculate the evidences, P(D|H1) and P(D|H2), where:
% 
%   P(D|Hi) = int{ P(D|w,Hi) * P(w|Hi) } dw  (Eq. 28.7) 
% 
%   D is [x] = [ -8 , -2 ,  6 ]
%        [t]   [  8 , 10 , 11 ] in MacKay's question
%
%   D can be user-defined in this extension
%
%   The measurement, t, is y + Normal(0,1)
% 
%   Model H1 is y = w0
%   Model H2 is y = w0 + w1*x
% 
%   Prior w0 is Normal(w0_E,w0_V) for both models    
%   Prior w1 is Normal(w1_E,w1_V)

% ---------------------------------------------------------------------------
% SOLUTION WITH (i) LAPLACE'S METHOD (Chapter 27) AND (ii) SYMBOLIC ALGEBRA 
% ---------------------------------------------------------------------------
% In Laplace's method, we find the peak of P(D|w,Hi) and Taylor expand to
% second order around this peak. We then approximate P(D|w,Hi) by a
% Gaussian and we approximate the normalizing constant of P(D|w,Hi) by the
% normalizing constant of this Gaussian. 
%
% In this question, P(D|w,Hi) is Gaussian, so the solution with Laplace's
% method will equal that found with symbolic algebra. This provides a useful
% check. Both are performed here. 
%
% The algebra is written for model H2; model H1 is a special case of H2.
% The algebra includes variables for expected values and variances so that
% they can be changed later. The measurement noise remains Normal(0,1).

% ---------------------------------------------------------------------------
% OBSERVATIONS
% ---------------------------------------------------------------------------
% Adding more data reduces the marginal likelihood (evidence) for both
% models but makes it easier to distinguish between the models. Only the ratio of
% marginal likelihoods is relevant for comparing models.
%
% The "log(margianal likelihood) per datapoint" is a better measure because
% it reduces the dependence on datapoints. 
%
% If the priors are badly wrong, the Evidence drops considerably, but you
% can make up for bad priors (to some extent) by increasing their variance.

% ---------------------------------------------------------------------------
% SET THE PLOT OPTIONS
% ---------------------------------------------------------------------------
%
% Set 'plot_options' { 0 = plot nothing , 1 = plot parameters , 2 = plot data }
plot_options = 2; 
% Start display
fprintf('----------------------------------------------\n')
fprintf('               w0       w1      log{ P(D|Hi) }\n')
fprintf('----------------------------------------------\n')

% ---------------------------------------------------------------------------
% SET THE DATA
% ---------------------------------------------------------------------------
%
method_data = 'from_model_H2';
switch method_data
    case 'MacKay_28_2'
        % Data from Exercise 28.2 in MacKay's book
        x = [ -8 , -2 ,  6 ];
        t = [  8 , 10 , 11 ];
    case 'user_defined'
        % Any user-defined data
        x = [-8,-2,6 ,-7,-1,5 ,-6,0,4 ]; % user-defined
        t = [ 7, 5,11, 8, 6,12, 8, 6,12]; % user-defined
    case 'from_model_H2'
        % Data taken from model H2 with measurement noise: y = w0 + w1*x + Normal(0,1) 
        w0_truth = 14;  % user-defined
        w1_truth = .1;  % user-defined
        x = -10:1:+10; % user-defined
        t = w0_truth + w1_truth*x + randn(size(x));
        fprintf('ground truth  %+6.4f  %+6.4f      - \n',w0_truth,w1_truth)
end

% ---------------------------------------------------------------------------
% SET THE PRIOR
% ---------------------------------------------------------------------------
%
method_prior = 'from_truth';
switch method_prior
    case 'MacKay_28_2'
        % Prior expected values and variances from MacKay 28.2
        w0_E = 0.0; w0_V = 1.0;
        w1_E = 0.0; w1_V = 1.0;
    case 'user_defined'
        % User-defined expected values and variances
        w0_E =  7.0; w0_V = 10;
        w1_E = -0.1; w1_V = 10;
    case 'from_truth'
        % Ground truth expected values and user-defined variances 
        w0_E = w0_truth; w0_V = 10;
        w1_E = w1_truth; w1_V = 10;
end
fprintf('prior         %+6.4f  %+6.4f      - \n',w0_E,w1_E)
% start the clock
tic

% ---------------------------------------------------------------------------
% Prior, Likelihood, and Posterior for Model H2
% ---------------------------------------------------------------------------
%
% 1. Consider Model H2
%
%   y = w0 + w1*x
%
% 2a. The Prior distribution for parameter w0 is:
%
%   p(w0|H2) = Normal(w0_E,w0_V)
%
%                     1              {   1 (w0_E - w0)^2 }
%   p(w0|H2) = --------------- * exp { - - ------------- }
%              sqrt(2*pi*w0_V)       {   2     w0_V      }
%
% 2b. The Prior distribution for parameter w1 is:
%
%   p(w1|H2) = Normal(w1_E,1)
%
%                     1              {   1 (w1_E - w1)^2 }
%   p(w1|H2) = --------------- * exp { - - ------------- }
%              sqrt(2*pi*w1_V)       {   2     w1_V      }
%
% 3. The Posterior distribution for w0,w1 is given by:
% 
%                   P(D|w0,w1,H2) * P(w0|H2) * P(w1|H2)
%   P(w0,w1|D,H2) = -----------------------------------       [1]
%                               P(D|H2)
%
% 4. The measurement, t, is y + epsilon, where epsilon is Normal(0,1):
%
%   t = y + epsilon
% 
%                                    1            {   1 (t-y(w0,w1,H2))^2 }
%   P(t|w0,w1,H2) = P(epsilon) = ---------- * exp { - - ----------------- }
%                                sqrt(2*pi)       {   2         1         }
% 
%   where y(w0,w1,H2) = w0 + w1*x for model H2
%
% 5. the likelihood is:
% 
%   P(D|w0,w1,H2) = P(t1|w0,w1,H2) * P(t2|w0,w1,H2) * P(t3|w0,w1,H2)
%
%                       1             {   1 (t1-w0-w1*x1)^2 + (t2-w0-w1*x2)^2 + (t3-w0-w1*x3)^2 }      . 
%              = -------------- * exp { - - --------------------------------------------------- } [2]  .
%                (sqrt(2*pi))^3       {   2                          1                          }      . 
%
% Note that the measurement variance is hard-wired to 1. User-defined
% measurement variance can be introduced by replacing 1 with t_V in the
% denominator of the argument and dividing the pre-exponential factor by
% sqrt(t_V). That change would then follow through the algebra below.
%
% 6. The Posterior distribution is:
%
%   P(w0,w1|D,H2)  =  P(D|w0,w1,H2) * P(w0|H2) * P(w1|H2)

% ---------------------------------------------------------------------------
% Find (w0,w1) that maximizes the Posterior distribution P(w0,w1|D,H2)
% ---------------------------------------------------------------------------
%
% We don't need to know the denominator of [1], P(D|H2), for this. We just
% maximize the numerator of the posterior distribution above: 
%
% Find w0,w1 that maximizes P(D|w0,w1,H2) * P(w0|H2) * P(w1|H2), which occurs when:
%
%       d   ( (t1-w0-w1*x1)^2 + (t2-w0-w1*x2)^2 + (t3-w0-w1*x3)^2   (w0_E-w0)^2   (w1_E-w1)^2 )
% 0 = ----- ( --------------------------------------------------- + ----------- + ----------- )
%     d(w0) (                          1                               w0_V           w1_V    )
%
% and
%
%       d   ( (t1-w0-w1*x1)^2 + (t2-w0-w1*x2)^2 + (t3-w0-w1*x3)^2   (w0_E-w0)^2   (w1_E-w1)^2 )
% 0 = ----- ( --------------------------------------------------- + ----------- + ----------- )
%     d(w1) (                          1                               w0_V           w1_V    )
%
% The first of these gives:
%
% 0 = -(t1-w0-w1*x1) - (t2-w0-w1*x2) - (t3-w0-w1*x3) - (w0_E-w0)/w0_V
% 0 = (3+1/W0_V)*w0 - (t1-w1*x1) - (t2-w1*x2) - (t3-w1*x3) - w0_E/w0_V
% 0 = (3+1/W0_V)*w0 + w1*(x1+x2+x3) - (t1+t2+t3) - w0_E/w0_V                       [1a]
%
% The second of these gives
%
% 0 = - 2*x1*(t1-w0-w1*x1) - 2*x2*(t2-w0-w1*x2) - 2*x3*(t3-w0-w1*x3) - 2*(w1_E-w1)/w1_V  
% 0 = - x1*(t1-w0-w1*x1) - x2*(t2-w0-w1*x2) - x3*(t3-w0-w1*x3) - (w1_E/w1_V-w1/w1_V)  
% 0 = w0(x1+x2+x3) + w1*(x1^2+x2^2+x3^2+1/w1_V) - (t1*x1+t2*x2+t3*x3) - w1_E/w1_V     [1b]
%
% Put into matrix form:
%  M * [w0] = [r0]          [1a]
%      [w1]   [r1]          [1b]
% LHS:
M_H2 = [ length(t)+1/w0_V      , sum(x)              ; ... % [LHS of 1a]
         sum(x)                , sum(x.^2) + 1/w1_V ];     % [LHS of 1b]
% RHS:
r_H2 = [ sum(t) + w0_E/w0_V     ; ... % [RHS of 1a]
         sum(t.*x) + w1_E/w1_V ];     % [RHS of 1b]
% Solve for w0 and w1 using linear algebra w = inv(M) * r
w = M_H2\r_H2;
% Extract w0 and w1 from w
w0_MAP_H2 = w(1); w1_MAP_H2 = w(2);

% ---------------------------------------------------------------------------
% Prior, Likelihood, and Posterior for Model H1
% ---------------------------------------------------------------------------
%
%   y = w0
%
% Find w0 that maximizes P(D|w0,H1) * P(w0|H1), which occurs when:
%
%       d   ( (t1-w0)^2 + (t2-w0)^2 + (t3-w0)^2   (w0_E-w0)^2 )
% 0 = ----- ( --------------------------------- + ----------- )
%     d(w0) (                       1                w0_V     )
w0_MAP_H1 = (sum(t) + w0_E/w0_V) / (length(t)+1/w0_V);

% ---------------------------------------------------------------------------
% Perform Taylor expansion of log(f[w0,w1]) around the MAP point for H2
% ---------------------------------------------------------------------------
%
% The numerator of the posterior f:= P(w0,w1|D,H2) in [1] is:
%
%                   1                    {   1 ( (t1-w0-w1*x1)^2 + (t2-w0-w1*x2)^2 + (t3-w0-w1*x3)^2   (w0_E-w0)^2   (w1_E-w1)^2 ) }  . 
% f = ---------------------------- * exp { - - ( --------------------------------------------------- + ----------- + ----------- ) }  .
%     sqrt((2*pi)^5 * w0_V * w1_V)       {   2 (                       1                                  w0_V           w1_V    ) }  . 
%
% Take the log of this:
%
% g = log(f) = -2.5*log(2*pi) - log(w0_V) - log(w1_V) - 0.5*(t1-w0-w1*x1)^2 + 0.5*(t2-w0-w1*x2)^2 + 0.5*(t3-w0-w1*x3)^2 + 0.5*(w0_E-w0)^2/w0_V + 0.5*(w1_E-w1)^2/w1_V 
%
% Perform a second order Taylor expansion on g:
%                                                                                          <------- Matrix A ------->  
%                        dg                dg                                              [ d2g/dw0dw0  d2g/dw0dw1 ]   [ delta(w0) ]                                     . 
% delta(g) = g[w0,w1]  + --- * delta(w0) + --- * delta(w1)  +  [ delta(w0) , delta(w1) ] * [                        ] * [           ]  +  [third and higher order terms]  . 
%                        dw0               dw1                                             [ d2g/dw1dw0  d2g/dw1dw1 ]   [ delta(w0) ]                                     .
%                        <------ first order terms ------>     <--------------------- second order terms --------------------------->                                     .
%
% dg
% --- = - (t1-w0-w1*x1) - (t2-w0-w1*x2) - (t3-w0-w1*x3) - (w0_E-w0)/w0_V
% dw0
%
%  d2g     
% ----- = length(t) + 1/w0_V
% dw0^2
%
% dg
% --- = - x1*(t1-w0-w1*x1) - x2*(t2-w0-w1*x2) - x3*(t3-w0-w1*x3) - (w1_E-w1)/w1_V
% dw1
%
%  d2g     
% ----- = x1^2 + x2^2 + x3^2 + 1/w1_V 
% dw1^2
%
%  d2g
% ------ = x1 + x2 + x3 
% dw0dw1
% 
% Note that, in this case, there are no third and higher order terms to the
% Taylor expansion of g. This is because g is quadratic, which is because
% the distributions that make up f are Gaussian. This is not true in general.  
%
% At the MAP point, dg/dw0 and dg/dw1 are zero (by definition of the MAP
% point). Therefore the first order terms in the Taylor expansion disappear
% and only the second order terms remain. 
%
% Around the MAP point, the posterior, f, takes the form:
%
%                 {   1                   [dw0] }
% f = f_max * exp { - - [dw0 , dw1] * A * [   ] }
%                 {   2                   [dw1] }
%
% where 
%   f_max is the peak value of f, occurring at (w0_max,w1_max)
%   dw0 = w0 - w0_max 
%   dw1 = w1 - w1_max 
% and
%
% A is the matrix labelled 'matrix A' above (repeated here for convenience):
%
%       [ d2g/dw0dw0  d2g/dw0dw1 ]
%   A = [                        ]
%       [ d2g/dw1dw0  d2g/dw1dw1 ]

% Perform the 2nd order Taylor expansion for H2
A_H2 = [ length(t) + 1/w0_V ,  sum(x)             ; ...
         sum(x)             ,  sum(x.^2) + 1/w1_V ];

% Perform the 2nd order Taylor expansion for H1 (by inspection)
A_H1 =   length(t) + 1/w0_V;

% ---------------------------------------------------------------------------
% Use Laplace's approximation to calculate the evidence for H2; from MacKay p350 equ (28.10)
% ---------------------------------------------------------------------------
%
% The integral of the posterior f, over the whole of parameter space (w0,w1) is approximated by: 
%
%                                              Matrix A written above 
%           <likelihood->   <--prior-->        |
% P(D|H2) ~ P(D|w_max,H2) * P(w_max|H2) * [det(A/2/pi)]^(-1/2)
%           <--- what I call f_max --->
%
%     f_max_H2 = f_max_H2 * (1/sqrt(2*pi) * exp( -(t(nn)-(w0_max_H2 + w1_max_H2*x(nn))).^2 ));
%
% logf_max is the value of logf at the peak, i.e. where w_max = [w0_max_H2,w1_max_H2]
logf_max_H2 = 0;
% Calculate the contribution of each datapoint to the log likelihood log{P(D|w_max,H2)}
for nn = 1:length(t)
    logf_max_H2 = logf_max_H2 - 0.5*log(2*pi) - 0.5 * ( t(nn) - ( w0_MAP_H2 + w1_MAP_H2*x(nn) ) ).^2 ;
end
% Calculate the contribution of the prior to logf_max
logf_max_H2 = logf_max_H2 - 0.5*log(2*pi*w0_V) - 0.5*log(2*pi*w1_V) - 0.5*(w0_E-w0_MAP_H2).^2/w0_V - 0.5*(w1_E-w1_MAP_H2).^2/w1_V;
% Estimate the logevidence with Laplace's method
log_P_D_H2 = logf_max_H2 - 0.5*log(det(A_H2/2/pi)); 

% ---------------------------------------------------------------------------
% Use Laplace's approximation to calculate the evidence for H1
% ---------------------------------------------------------------------------
%
logf_max_H1 = 0;
% Calculate the contribution of each datapoint to the log likelihood log{P(D|w_max,H1)}
for nn = 1:length(t)
    logf_max_H1 = logf_max_H1 - 0.5*log(2*pi) - 0.5 * ( t(nn) - w0_MAP_H1 ).^2 ;
end
% Calculate the contribution of the prior to logf_max
logf_max_H1 = logf_max_H1 - 0.5*log(2*pi*w0_V) - 0.5*(w0_E-w0_MAP_H1)^2/w0_V;
% Estimate the logevidence with Laplace's method
log_P_D_H1 = logf_max_H1 - 0.5*log(A_H1/2/pi);
% Stop the clock
time_Laplace = toc;

% Display
fprintf('Using Laplace''s method: \n')
fprintf('Model H1      %+6.4f    -     %10.5f \n',w0_MAP_H1,log_P_D_H1)
fprintf('Model H2      %+6.4f  %+6.4f %10.5f \n',w0_MAP_H2,w1_MAP_H2,log_P_D_H2)

%---------------------------------------------------------
% USE SYMBOLIC ALGEBRA IF THERE ARE NOT MANY DATAPOINTS
%---------------------------------------------------------
if length(x) <= 10
    tic

    %% Create expressions for the posterior likelihood
    %
    % Define two symbolic variables
    syms w0_sys w1_sys
    % Set the prior likelihood P(w|H_i) for models H1 and H2 (label this J)
    J_sys_H1 = (2*pi*w0_V)^(-0.5)*exp(-0.5*(w0_sys-w0_E)^2/w0_V);
    J_sys_H2 = (2*pi*w0_V)^(-0.5)*exp(-0.5*(w0_sys-w0_E)^2/w0_V) * (2*pi*w1_V)^(-0.5)*exp(-0.5*(w1_sys-w1_E)^2/w1_V);
    % Assimilate the data by adding the likelihood P(D|w,H_i) to each model for each datapoint in turn
    for n = 1:length(x)
        J_sys_H1 = J_sys_H1 * (2*pi)^(-0.5)*exp(-0.5*(t(n)-w0_sys)^2); % n.b. variance hard-wired to 1
        J_sys_H2 = J_sys_H2 * (2*pi)^(-0.5)*exp(-0.5*(t(n)-w0_sys-w1_sys*x(n))^2); % n.b. variance hard-wired to 1
    end

    %% Calculate the Marginal Likelihood (without finding the MAP point
    %
    % Create the integral of P(D|w,H_i) w.r.t. w
    ML_sys_H1 = int(J_sys_H1, w0_sys, -inf,+inf);
    ML_sys_H2 = int(int(J_sys_H2, w0_sys, -inf, +inf), w1_sys, -inf,+inf);
    % Evaluate the Marginal Likelihoods
    logevi_Laplace_sys_H1 = log(eval(ML_sys_H1));
    logevi_Laplace_sys_H2 = log(eval(ML_sys_H2));

    %% Find the MAP points
    %
    % Create the partial derivatives
    dJ_sys_H1_dw0 = diff(J_sys_H1, w0_sys);
    dJ_sys_H2_dw0 = diff(J_sys_H2, w0_sys);
    dJ_sys_H2_dw1 = diff(J_sys_H2, w1_sys);
    % Find the MAP point for H1
    w0_sys_MAP_H1 = vpa(solve(dJ_sys_H1_dw0 == 0));
    % Find the MAP point for H2
    wn_sys_MAP_H2 = solve([dJ_sys_H2_dw0 == 0, dJ_sys_H2_dw1 == 0], [w0_sys w1_sys]);
    w0_sys_MAP_H2 = vpa(wn_sys_MAP_H2.w0_sys);
    w1_sys_MAP_H2 = vpa(wn_sys_MAP_H2.w1_sys);
    % Stop the clock
    time_symbolic = toc;
    fprintf('--------------------------------------------- \n')
    fprintf('Using symbolic algebra: \n')
    fprintf('Model H1      %+6.4f    -     %10.5f \n',w0_sys_MAP_H1,logevi_Laplace_sys_H1)
    fprintf('Model H2      %+6.4f  %+6.4f %10.5f \n',w0_sys_MAP_H2,w1_sys_MAP_H2,logevi_Laplace_sys_H2)
else
    time_symbolic = NaN;
end

fprintf('------------------------------------------ \n')
fprintf('Evidence for H1 / Evidence for H2: %6.4g \n',exp(log_P_D_H1 - log_P_D_H2))
fprintf('--------------------------------------------- \n')
if plot_options == 0
    fprintf('Time for Laplace''s method:  %+6.4f s \n',time_Laplace)
    fprintf('Time for symbolic algebra :  %+6.4f s \n',time_symbolic)
else
    fprintf('Set verbose = 0 to compare times taken   \n')
end

%---------------------------------------------------------
% Plot data and models
%---------------------------------------------------------
if plot_options == 1
    figure(1); clf
    % Set the abscissa for w0 and w1
    N = 51; w0 = linspace(4.5,10,N); w1 = linspace(-0.4,0.6,N);
    % mesh this
    [w0w0,w1w1] = meshgrid(w0,w1);
    % Calculate the log likelihood P(D|w0,w1,H2)
    loglik_H2 = zeros(N,N);
    for nn = 1:length(t)
        loglik_H2 = loglik_H2 - 0.5*log(2*pi) - 0.5*( t(nn) - (w0w0 + w1w1*x(nn)) ).^2 ;
        % Calculate the log prior P(w0|H2)
        logprior_H2 = - log(2*pi) - 0.5*log(w0_V) - 0.5*log(w1_V) - 0.5*(w0_E - w0w0).^2/w0_V - 0.5*(w1_E - w1w1).^2/w1_V;
        [~,han_brute] = contourf(w0w0,w1w1,(loglik_H2 + logprior_H2),50); shading flat
        hold on
        han_calc = plot(w0_MAP_H2,w1_MAP_H2,'ro','MarkerSize',20);
        xlabel('w0')
        ylabel('w1')
        colorbar
        title('log(posterior likelihood)')
        legend([han_brute,han_calc],{'point-wise evaluation','minimize gradient by hand calculation'})
    end
elseif plot_options == 2
    figure(2); clf
    hold on
    han_data = errorbar(x,t,2*sqrt(1)*ones(1,length(t)),'ko','MarkerFaceColor','k');
    han_H1   = plot([min(x),max(x)],[w0_MAP_H1,w0_MAP_H1],'r-');
    han_H2   = plot([min(x),max(x)],[w0_MAP_H2+w1_MAP_H2*min(x),w0_MAP_H2+w1_MAP_H2*max(x)],'g-');
    legend([han_data,han_H1,han_H2],{'data ± 2 s.d.','model H1','model H2'})
    xlabel('x'); ylabel('y'); title('results vs. model predictions (n.b. model includes priors on model parameters)','Interpreter','none')
    grid on
    xlim([min(x),max(x)])
end

end
